/* Talika panel applet: Switch between open windows using icons.
 * (C) 2009-2010 Sinew Software Systems
	 *
 * Authors:  Vinod Kumar Dahiya
 *           Hemant Kumar
 *
 * NOTE : A lot of code ( about 90% ) has been taken from libwnck tasklist.c
	 * we just renamed & modified it to suit our purpose.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
	 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include <gtk/gtk.h>
#include <libwnck/libwnck.h>
#include <math.h>

#include "tasklist.h"
#include "task.h"
#include "xutils.h"
#include "tooltips.h"
#include "launcher.h"
#include "button-widget.h"

static void 
talika_task_stop_glow (TalikaTask * task);

static void
talika_task_popup_menu (TalikaTask * task, gboolean action_submenu);

static gboolean
talika_task_get_needs_attention (TalikaTask * task);

/*########################### Drag and drop ##################################*/

static void
talika_task_get_drop_popup_xy (TalikaTask * task, 
                              gint * x, gint * y,
                              gint yPadding)
{
	GtkWidget *widget = task->drop_popup;
	gint panel_size;
	GtkRequisition requisition;
	gtk_widget_size_request (widget, &requisition);

	gdk_window_get_origin (task->button->window, x, y);

	*x += task->button->allocation.x;
	*y += task->button->allocation.y;

	panel_size = task->tasklist->panel_size;

	if (task->tasklist->panel_orient == PANEL_APPLET_ORIENT_UP)
	{
		*y = *y - requisition.height - yPadding;
	}
	else if (task->tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
	{
		*y = *y + panel_size + yPadding;
	}
	else if (task->tasklist->panel_orient == PANEL_APPLET_ORIENT_RIGHT)
	{
		*x = *x + panel_size + yPadding;
	}
	else
	{
		*x = *x - requisition.width - yPadding;
	}
}

static void
talika_task_position_drop_popup (TalikaTask * task)
{
	gint x, y;
	GtkRequisition event_box_req;
	GtkRequisition popup_req;

	gtk_widget_size_request (task->button, &event_box_req);
	gtk_widget_size_request (task->drop_popup, &popup_req);

	talika_task_get_drop_popup_xy (task, &x, &y, 0);

	GdkScreen *screen = gtk_widget_get_screen (task->button);
	GdkRectangle monitor_rect;
	gdk_screen_get_monitor_geometry (screen,
	                                 gdk_screen_get_monitor_at_point
	                                 (screen, x, y), &monitor_rect);
	gint monitor_right = monitor_rect.x + monitor_rect.width;
	gint monitor_left = monitor_rect.x;
	gint monitor_bottom = monitor_rect.y + monitor_rect.height;
	gint monitor_top = monitor_rect.y;


	if (task->tasklist->panel_orient == PANEL_APPLET_ORIENT_UP ||
	    task->tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
	{
		//x = x - (popup_req.width / 2) + (event_box_req.width / 2);

		if (x + popup_req.width >= monitor_right - 5)
		{
			x = monitor_right - popup_req.width - 5;
		}
		else if (x < monitor_left + 5)
		{
			x = monitor_left + 5;
		}
	}
	else
	{
		//y = y - (popup_req.height / 2) + (event_box_req.height / 2);

		if (y + popup_req.height >= monitor_bottom - 5)
		{
			y = monitor_bottom - popup_req.height - 5;
		}
		else if (y < monitor_top + 5)
		{
			y = monitor_top + 5;
		}
	}

	gtk_window_move (GTK_WINDOW (task->drop_popup), x, y);
}

static void
talika_task_show_drop_popup (TalikaTask * task)
{
	if (task->drop_popup)
		return;

	//create menu
	GtkWidget *menu;
	TalikaTask *win_task;
	char *text;
	GdkPixbuf *pixbuf;
	GtkWidget *menu_item;
	GtkWidget *image;
	GtkWidget *frame;
	GtkWidget *button;
	//GtkWidget *hbox;
	GList *l;

	g_return_if_fail (task->type == TALIKA_TASK_CLASS_GROUP);

	if (task->class_group == NULL)
		return;

	//menu = gtk_menu_new ();
	menu = gtk_vbox_new (TRUE,0);

	l = task->windows;
	while (l)
	{
		win_task = l->data;

		menu_item = gtk_hbox_new (FALSE,0);

		pixbuf = gdk_pixbuf_copy (wnck_window_get_mini_icon
		                          (win_task->window));
		if (pixbuf)
		{
			image = gtk_image_new_from_pixbuf (pixbuf);
			gtk_widget_show (image);
			gtk_box_pack_start (GTK_BOX(menu_item),image,
		                  FALSE,FALSE,0);
			g_object_unref (pixbuf);
		}
		
		text = talika_task_get_text (win_task);
		button = gtk_menu_item_new_with_label (text);
		g_free (text);

		if (talika_task_get_needs_attention (win_task))
			_make_gtk_label_bold (GTK_LABEL
			                      (GTK_BIN (button)->child));

		text = talika_task_get_text (win_task);
		gtk_widget_set_tooltip_text (button, text);
		g_free (text);

		//gtk_button_set_relief (GTK_BUTTON(button),GTK_RELIEF_NONE);
		gtk_box_pack_start (GTK_BOX(menu_item),button,
		                  FALSE,FALSE,0);

		//drag drop connect
		static const GtkTargetEntry targets[] = {
			{"application/x-wnck-window-id", 0, 0}
		};

		gtk_drag_dest_set (GTK_WIDGET (button),
	                   GTK_DEST_DEFAULT_DROP, targets, 1,
	                   GDK_ACTION_MOVE);

		g_signal_connect (G_OBJECT (button), "drag_motion",
	                  G_CALLBACK (talika_task_drag_motion), win_task);

		g_signal_connect (G_OBJECT (button), "drag_leave",
	                  G_CALLBACK (talika_task_drag_leave), win_task);

		gtk_box_pack_start_defaults (GTK_BOX(menu), menu_item);
		
		l = l->next;
	}

	//create popup window
	task->drop_popup = gtk_window_new (GTK_WINDOW_POPUP);
	gtk_window_resize (GTK_WINDOW (task->drop_popup), 10,
		                   10);
	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
	
	gtk_container_add (GTK_CONTAINER(task->drop_popup),frame);
	gtk_container_add (GTK_CONTAINER(frame),menu);
	gtk_container_set_border_width (GTK_CONTAINER(menu),2);

	gtk_widget_show_all (task->drop_popup);
	talika_task_position_drop_popup (task);
}

void
talika_task_set_sort_order (TalikaTask *task, gint sort_order)
{
	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
		{
			TalikaTask *temp;
			temp = g_list_last (task->windows)->data;
			wnck_window_set_sort_order (temp->window, sort_order);
		}
		break;
			
		case TALIKA_TASK_WINDOW:
		{
			wnck_window_set_sort_order (task->window, sort_order);
		}
		break;

		case TALIKA_TASK_LAUNCHER:
		{
			task->sort_order = sort_order;
		}
		break;
	}
}

gint
talika_task_get_sort_order (TalikaTask *task)
{
	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
		{
			TalikaTask *temp;
			temp = g_list_last (task->windows)->data;
			return wnck_window_get_sort_order (temp->window);
		}
		break;
			
		case TALIKA_TASK_WINDOW:
		{
			return wnck_window_get_sort_order (task->window);
		}
		break;

		case TALIKA_TASK_LAUNCHER:
		{
			return task->sort_order;
		}
		break;
	}

	//never reach here
	return 0;
}

static gboolean 
talika_task_drop_popup_timeout (gpointer data)
{
	TalikaTask *task = data;

	task->drop_popup_activate = 0;

	if(task->type == TALIKA_TASK_CLASS_GROUP)
	{
		if (task->drop_popup);
		{
			gtk_widget_destroy (task->drop_popup);
			task->drop_popup = NULL;
		}
	}

	task->dnd_timestamp = 0;

	return FALSE;
}

static gboolean 
talika_task_motion_timeout (gpointer data)
{
	WnckWorkspace *ws;
	TalikaTask *task = data;

	task->button_activate = 0;

	if(task->type == TALIKA_TASK_WINDOW)
	{
		/* FIXME: THIS IS SICK AND WRONG AND BUGGY.  See the end of
		 * http://mail.gnome.org/archives/wm-spec-list/2005-July/msg00032.html
		 * There should only be *one* activate call.
		 */
		ws = wnck_window_get_workspace (task->window);
		if (ws
		    && ws !=
		    wnck_screen_get_active_workspace (wnck_screen_get_default ()))
		{
			wnck_workspace_activate (ws, task->dnd_timestamp);
		}
		wnck_window_activate_transient (task->window, task->dnd_timestamp);
	}
	else if (task->type == TALIKA_TASK_CLASS_GROUP)
	{
		talika_task_show_drop_popup (task);
	}
	else if (task->type == TALIKA_TASK_LAUNCHER)
	{
		//talika_launcher_launch (task->button,task);
	}

	task->dnd_timestamp = 0;

	return FALSE;
}

void 
talika_task_drag_leave (GtkWidget * widget,
                        GdkDragContext * context,
                        guint time, TalikaTask * task)
{
	if (task->button_activate != 0)
	{
		g_source_remove (task->button_activate);
		task->button_activate = 0;
	}

	gtk_drag_unhighlight (widget);

	//start popup hiding if task was a class task or it is in a class task
	if (task->tasklist->use_grouping  && task->type != TALIKA_TASK_LAUNCHER)
	{
		TalikaTask *class_task;
		if (task->type == TALIKA_TASK_CLASS_GROUP)
		{
			class_task = task;
		}
		else
		{
			class_task = g_hash_table_lookup (
			                                  task->tasklist->class_group_hash,
			                                  task->class_group_name);
		}

		if (class_task&&class_task->drop_popup)
		{
			if (class_task->drop_popup_activate == 0)
				class_task->drop_popup_activate =
				g_timeout_add (WNCK_ACTIVATE_TIMEOUT,
				               talika_task_drop_popup_timeout,
				               class_task);
		}
	}
}

gboolean 
talika_task_drag_motion (GtkWidget * widget,
                         GdkDragContext * context,
                         gint x, gint y, 
                         guint time, 
                         TalikaTask * task)
{
	//cancel popup hiding if task is a class task or it is in a class task
	if (task->tasklist->use_grouping && task->type != TALIKA_TASK_LAUNCHER)
	{
		TalikaTask *class_task;
		if (task->type == TALIKA_TASK_CLASS_GROUP)
		{
			class_task = task;
		}
		else
		{
			class_task = g_hash_table_lookup (
			                                  task->tasklist->class_group_hash,
			                                  task->class_group_name);
		}

		if (class_task->drop_popup_activate != 0)
		{
			g_source_remove (class_task->drop_popup_activate);
			class_task->drop_popup_activate = 0;
		}
	}

	//do the rest
	if (gtk_drag_dest_find_target (widget, context, NULL))
	{
		gtk_drag_highlight (widget);
		gdk_drag_status (context, context->suggested_action, time);
	}
	else
	{
		gdk_drag_status (context, 0, time);
	}

	//activate on timeout
	task->dnd_timestamp = time;
	if (task->button_activate == 0)
		task->button_activate =
		g_timeout_add (WNCK_ACTIVATE_TIMEOUT,
		               talika_task_motion_timeout,
		               task);
	
	return TRUE;
}

void 
talika_task_drag_begin (GtkWidget * widget,
                        GdkDragContext * context, TalikaTask * task)
{
	GdkPixbuf *buf = talika_task_get_icon (task);
	gtk_drag_set_icon_pixbuf (context, buf, -2, -2);
	task->tasklist->drag_start_time = gtk_get_current_event_time ();
}

void
talika_task_drag_end (GtkWidget * widget,
                      GdkDragContext * context, TalikaTask * task)
{
	task->tasklist->drag_start_time = 0;
}

void
talika_task_drag_data_get (GtkWidget * widget,
                           GdkDragContext * context,
                           GtkSelectionData * selection_data,
                           guint info, guint time, TalikaTask * task)
{
	TalikaTaskList *tasklist = task->tasklist;
	gint nth = g_list_index (tasklist->visible_tasks, task);

	gtk_selection_data_set (selection_data,
	                        selection_data->target,
	                        8, (guchar *) & nth, sizeof (gint));
}

void
talika_task_drag_data_received (GtkWidget * widget,
                                GdkDragContext * context,
                                gint x,
                                gint y,
                                GtkSelectionData * data,
                                guint info,
                                guint time, TalikaTask * target_task)
{
	TalikaTaskList *tasklist;
	TalikaTask *source_task = NULL;
	
	tasklist = target_task->tasklist;
	
	switch (info)
	{
		case 1:
		{
			gboolean app_found = FALSE;
			gchar **uris = g_uri_list_extract_uris ((char *)data->data);

			//see if first file is a desktop file
			if (uris[0])
			{
				gchar *test = uris[0];
				int len = strlen (test);
				if (g_ascii_strcasecmp (test+len-8,".desktop") == 0)
				{
					//get desktop identifier
					gchar *desktop_id = g_path_get_basename (test);
					int dlen = strlen (desktop_id);
					desktop_id[dlen-8]='\0';
					strcat(desktop_id,":nil");

					//create launcher
					TalikaTask *launcher;
					launcher = talika_launcher_new (tasklist,desktop_id);

					if(launcher)
					{
						talika_task_set_sort_order (launcher,-100);
						tasklist->launchers = g_list_append 
							(tasklist->launchers,launcher);
						talika_tasklist_save_launchers (tasklist);
						
						app_found = TRUE;
					}

					g_free (desktop_id);

					//assign as source task
					source_task = launcher;
				}
			}

			//otherwise launch if required
			if (!app_found)
			{
				GList *list = NULL;

				int i;
				for (i = 0; uris[i]; i++) 
				{
					list = g_list_append (list,uris[i]);
				}

				if (target_task->type == TALIKA_TASK_LAUNCHER && list)
				{
					talika_launcher_launch_uris (target_task->button,target_task,list);
					gtk_drag_finish (context, TRUE, FALSE, time);
				}
				else
				{
					gtk_drag_finish (context, FALSE, FALSE, time);
				}

				g_list_free (list);
				return;
			}
		}
		break;

		case 0:
		{
			//check if data is of our concern
			if ((data->length != sizeof (gint)) || (data->format != 8))
			{
				gtk_drag_finish (context, FALSE, FALSE, time);
				return;
			}

			//get source task
			gint *nth = (gint *) data->data;
			source_task = g_list_nth_data (tasklist->visible_tasks, *nth);

			//if task is a window task, and drag time is too short on same task
			if (source_task->type == TALIKA_TASK_WINDOW)
			{
				if (target_task == source_task)
				{
					GtkSettings *settings;
					gint double_click_time;

					settings =
						gtk_settings_get_for_screen (gtk_widget_get_screen
						                             (GTK_WIDGET (tasklist->box)));
					double_click_time = 0;
					g_object_get (G_OBJECT (settings),
					              "gtk-double-click-time", &double_click_time,
					              NULL);

					if ((time - tasklist->drag_start_time) < double_click_time)
					{
						talika_tasklist_activate_task_window (target_task,
						                                      time,FALSE);
						gtk_drag_finish (context, TRUE, FALSE, time);
						return;
					}
				}
			}
		}
		break;

		default:
			return;
	}

	//update order of tasks now
	GList *l, *windows;
	WnckWindow *window;
	guint new_order, old_order, order;


	//determine new order of task
	old_order = talika_task_get_sort_order (source_task);
	new_order = talika_task_get_sort_order (target_task);
	if (old_order < new_order)
		new_order++;

	//update launchers first
	GList *ls;
	for (ls = tasklist->launchers; ls; ls = ls->next)
	{
		TalikaTask *launcher = ls->data;

		order = talika_task_get_sort_order (launcher);
		if (order >= new_order)
		{
			talika_task_set_sort_order (launcher, order + 1);
		}
	}
	talika_tasklist_save_launchers (tasklist);

	//update windows
	windows = wnck_screen_get_windows (tasklist->screen);
	for (l = windows; l; l = l->next)
	{
		window = WNCK_WINDOW (l->data);
		order = wnck_window_get_sort_order (window);
		if (order >= new_order)
			wnck_window_set_sort_order (window,
			                            order + 1);
	}

	//set sort order of source task
	talika_task_set_sort_order (source_task, new_order);
	if(source_task->launcher 
	   && (source_task->launcher->attached_task == source_task))
	{
		talika_task_set_sort_order (source_task->launcher, new_order);
	}

	//update list and finish drag
	talika_tasklist_update_lists (tasklist);
	//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));

	gtk_drag_finish (context, TRUE, FALSE, time);
}



/*##################### Attention demanding task glowing #####################*/

static gboolean
talika_task_get_needs_attention (TalikaTask * task)
{
	GList *l;
	TalikaTask *win_task;
	gboolean needs_attention;

	needs_attention = FALSE;

	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
		{
			task->start_needs_attention = 0;
			l = task->windows;
			while (l)
			{
				win_task = (TalikaTask *) l->data;

				if (wnck_window_or_transient_needs_attention
				    (win_task->window))
				{
					needs_attention = TRUE;
					time (&(task->start_needs_attention));
					break;
				}

				l = l->next;
			}
		}
			break;

		case TALIKA_TASK_WINDOW:
		{
			needs_attention =
				wnck_window_or_transient_needs_attention
				(task->window);
			if (needs_attention)
				time (&(task->start_needs_attention));
		}
			break;
	}

	return needs_attention != FALSE;
}

static gboolean
talika_task_button_glow (TalikaTask * task)
{
	GTimeVal tv;
	gdouble glow_factor, now;
	gfloat fade_opacity, loop_time;
	gint fade_max_loops;
	gboolean stopped;
	cairo_t *cr;

	if (task->screenshot == NULL)
		return TRUE;

	g_get_current_time (&tv);
	now = (tv.tv_sec * (1.0 * G_USEC_PER_SEC) +
	       tv.tv_usec) / G_USEC_PER_SEC;

	if (task->glow_start_time <= G_MINDOUBLE)
		task->glow_start_time = now;

	fade_opacity = 0.8;
	fade_max_loops = 5;
	loop_time = 3;

	if (task->button_glow == 0)
	{
		/* we're in "has stopped glowing" mode */
		glow_factor = 0;	//fade_opacity * 0.5;
		stopped = TRUE;
	}
	else
	{
		glow_factor = fade_opacity * (0.5 -
		                              0.5 *
		                              cos ((now -
		                                    task->glow_start_time) *
		                                   M_PI * 2.0 / loop_time));

		if (now - task->start_needs_attention >
		    loop_time * 1.0 * fade_max_loops)
			stopped =
			ABS (glow_factor - fade_opacity * 0.5) < 0.05;
		else
			stopped = FALSE;
	}

	gdk_window_begin_paint_rect (task->button->window,
	                             &task->button->allocation);

	cr = gdk_cairo_create (task->button->window);
	gdk_cairo_rectangle (cr, &task->button->allocation);
	cairo_translate (cr, task->button->allocation.x,
	                 task->button->allocation.y);
	cairo_clip (cr);

	cairo_save (cr);

	gdk_cairo_set_source_pixmap (cr, task->screenshot, 0., 0.);
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint (cr);

	cairo_restore (cr);

	gdk_cairo_set_source_pixmap (cr, task->screenshot_faded, 0., 0.);
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_paint_with_alpha (cr, glow_factor);

	cairo_destroy (cr);

	gdk_window_end_paint (task->button->window);

	if (stopped)
		talika_task_stop_glow (task);

	return !stopped;
}

static void
talika_task_clear_glow_start_timeout_id (TalikaTask * task)
{
	task->button_glow = 0;
}

static void
talika_task_queue_glow (TalikaTask * task)
{
	if (task->button_glow == 0)
	{
		task->glow_start_time = 0.0;

		/* The animation doesn't speed up or slow down based on the
		 * timeout value, but instead will just appear smoother or 
		 * choppier.
		 */
		task->button_glow =
			g_timeout_add_full (G_PRIORITY_DEFAULT_IDLE,
			                    50,
			                    (GSourceFunc)
			                    talika_task_button_glow, task,
			                    (GDestroyNotify)
			                    talika_task_clear_glow_start_timeout_id);
	}
}

static void 
talika_task_stop_glow (TalikaTask * task)
{
	if (task->button_glow != 0)
		g_source_remove (task->button_glow);
}

static GdkPixmap *
take_screenshot (TalikaTask * task)
{
	TalikaTaskList *tasklist;
	GtkWidget *tasklist_widget;
	GdkPixmap *pixmap;
	gint width, height;
	gboolean overlay_rect = TRUE;

	width = task->button->allocation.width;
	height = task->button->allocation.height;

	pixmap = gdk_pixmap_new (task->button->window, width, height, -1);

	tasklist = task->tasklist;
	tasklist_widget = GTK_WIDGET (task->tasklist->box);

	/* first draw the button */
	//gtk_widget_style_get (tasklist_widget, "fade-overlay-rect", &overlay_rect, NULL);
	if (overlay_rect)
	{
		/* Draw a rectangle with bg[SELECTED] */
		gdk_draw_rectangle (pixmap,
		                    task->button->style->
		                    bg_gc[GTK_STATE_SELECTED], TRUE, 0, 0,
		                    width + 1, height + 1);
	}
	else
	{
		GtkStyle *style;
		GtkStyle *attached_style;

		/* copy the style to change its colors around. */
		style = gtk_style_copy (task->button->style);
		style->bg[task->button->state] =
			style->bg[GTK_STATE_SELECTED];
		/* Now attach it to the window */
		attached_style = gtk_style_attach (style, pixmap);
		g_object_ref (attached_style);

		/* copy the background */
		gdk_draw_drawable (pixmap,
		                   attached_style->bg_gc[GTK_STATE_NORMAL],
		                   tasklist->background,
		                   task->button->allocation.x,
		                   task->button->allocation.y, 0, 0, width,
		                   height);

		/* draw the button with our modified style instead of the real one. */
		gtk_paint_box (attached_style, (GdkWindow *) pixmap,
		               task->button->state, GTK_SHADOW_OUT, NULL,
		               task->button, "button", 0, 0, width, height);

		g_object_unref (style);
		gtk_style_detach (attached_style);
		g_object_unref (attached_style);
	}

	/* expose the button image on pixmap*/
	GtkWidget *widget = task->button;
	GdkPixbuf *pb = button_widget_get_pixbuf (BUTTON_WIDGET(widget));

	int x, y, w, h;
	
	w = gdk_pixbuf_get_width (pb);
	h = gdk_pixbuf_get_height (pb);
	x = (widget->allocation.width - w)/2;
	y = (widget->allocation.height - h)/2;
		
	gdk_draw_pixbuf ((GdkWindow *) pixmap, NULL, pb,
				 0, 0,
				 x, y,
				 w, h,
				 GDK_RGB_DITHER_NORMAL,
				 0, 0);
	
	g_object_unref (pb);

	
	return pixmap;
}

static GdkPixmap *
copy_pixmap (GtkWidget * widget)
{
	GdkPixmap *pixmap;

	pixmap = gdk_pixmap_new (widget->window,
	                         widget->allocation.width,
	                         widget->allocation.height, -1);

	gdk_draw_drawable (pixmap,
	                   widget->style->bg_gc[GTK_STATE_NORMAL],
	                   widget->window,
	                   widget->allocation.x, widget->allocation.y,
	                   0, 0,
	                   widget->allocation.width,
	                   widget->allocation.height);

	return pixmap;
}


static void
cleanup_screenshots (TalikaTask * task)
{
	if (task->screenshot != NULL)
	{
		g_object_unref (task->screenshot);
		task->screenshot = NULL;
	}
	if (task->screenshot_faded != NULL)
	{
		g_object_unref (task->screenshot_faded);
		task->screenshot_faded = NULL;
	}
}

static gboolean
talika_task_expose (GtkWidget * widget, GdkEventExpose * event, gpointer data)
{
	GtkStyle *style;
	GdkGC *lgc, *dgc;
	int x, y, w;
	TalikaTask *task;

	task = (TalikaTask *) data;

	cleanup_screenshots (task);

	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
		{
			if(!task->tasklist->running_indicator)
			{
				style = widget->style;
				lgc = style->light_gc[GTK_STATE_NORMAL];
				dgc = style->dark_gc[GTK_STATE_NORMAL];

				x = widget->allocation.x + widget->allocation.width -
					(GTK_CONTAINER (widget)->border_width +
					 style->ythickness + 12);
				y = widget->allocation.y + widget->allocation.height / 2 - 5;

				//////////////////////////////// DRAW ARROW //////////////////////////////////
				GtkArrowType arrow_type = GTK_ARROW_UP;

				w = (widget->allocation.width -
				     GTK_CONTAINER (widget)->border_width -
				     style->ythickness * 2) / 4;

				switch (task->tasklist->panel_orient)
				{
					case PANEL_APPLET_ORIENT_UP:
					{
						x = widget->allocation.x + widget->allocation.width -
							(GTK_CONTAINER (widget)->border_width +
							 style->ythickness + w);

						y = widget->allocation.y +
							GTK_CONTAINER (widget)->border_width +
							style->ythickness;

						arrow_type = GTK_ARROW_UP;
					}
						break;

					case PANEL_APPLET_ORIENT_DOWN:
					{
						x = widget->allocation.x + widget->allocation.width -
							(GTK_CONTAINER (widget)->border_width +
							 style->ythickness + w);

						y = widget->allocation.y + widget->allocation.height -
							(GTK_CONTAINER (widget)->border_width +
							 style->ythickness + w);

						arrow_type = GTK_ARROW_DOWN;
					}
						break;

					case PANEL_APPLET_ORIENT_RIGHT:
					{
						x = widget->allocation.x + widget->allocation.width -
							(GTK_CONTAINER (widget)->border_width +
							 style->ythickness + w);

						y = widget->allocation.y + widget->allocation.height -
							(GTK_CONTAINER (widget)->border_width +
							 style->ythickness + w);

						arrow_type = GTK_ARROW_RIGHT;
					}
						break;

					case PANEL_APPLET_ORIENT_LEFT:
					{
						x = widget->allocation.x +
							GTK_CONTAINER (widget)->border_width +
							style->ythickness;

						y = widget->allocation.y + widget->allocation.height -
							(GTK_CONTAINER (widget)->border_width +
							 style->ythickness + w);

						arrow_type = GTK_ARROW_LEFT;
					}
						break;

					default:
						break;
				}

				gtk_paint_arrow (widget->style, widget->window,
				                 task->tasklist->active_class_group ==
				                 task ? GTK_STATE_ACTIVE : GTK_STATE_NORMAL,
				                 GTK_SHADOW_OUT, NULL, widget, NULL,
				                 arrow_type, TRUE, x, y, w, w);
			}
		}

			//////////////////////////////////////////////////////////////////////////////

			/* Fall through to get screenshot
			 */
		case TALIKA_TASK_WINDOW:
		{
			if ((event->area.x <= widget->allocation.x) &&
			    (event->area.y <= widget->allocation.y) &&
			    (event->area.width >= widget->allocation.width) &&
			    (event->area.height >= widget->allocation.height))
			{
				if (task->start_needs_attention)
				{
					task->screenshot = copy_pixmap (widget);
					task->screenshot_faded =
						take_screenshot (task);

					talika_task_button_glow (task);
				}
			}
		}
			break;
	}

	return FALSE;
}

/*################ Button press and Corrsponding Menu handling ###############*/

static void
talika_task_position_menu (GtkMenu * menu,
                           int *x,
                           int *y, gboolean * push_in, gpointer user_data)
{
	TalikaTask *task = (TalikaTask *) user_data;
	GtkWidget *widget = task->button;

	GtkRequisition requisition;
	GdkScreen *screen;
	int menu_x = 0;
	int menu_y = 0;
	int pointer_x;
	int pointer_y;

	screen = gtk_widget_get_screen (widget);

	gtk_widget_size_request (GTK_WIDGET (menu), &requisition);

	gdk_window_get_origin (widget->window, &menu_x, &menu_y);
	gtk_widget_get_pointer (widget, &pointer_x, &pointer_y);

	menu_x += widget->allocation.x;
	menu_y += widget->allocation.y;

	if (task->tasklist->panel_orient == PANEL_APPLET_ORIENT_UP ||
	    task->tasklist->panel_orient == PANEL_APPLET_ORIENT_DOWN)
	{
		if (gtk_widget_get_direction (GTK_WIDGET (menu)) !=
		    GTK_TEXT_DIR_RTL)
		{
			if (pointer_x < widget->allocation.width &&
			    requisition.width < pointer_x)
				menu_x += MIN (pointer_x,
				               widget->allocation.width -
				               requisition.width);
		}
		else
		{
			menu_x +=
				widget->allocation.width - requisition.width;
			if (pointer_x > 0
			    && pointer_x < widget->allocation.width
			    && pointer_x <
			    widget->allocation.width - requisition.width)
			{
				menu_x -=
					MIN (widget->allocation.width -
					     pointer_x,
					     widget->allocation.width -
					     requisition.width);
			}
		}
		menu_x = MIN (menu_x,
		              gdk_screen_get_width (screen) -
		              requisition.width);

		if (menu_y > gdk_screen_get_height (screen) / 2)
			menu_y -= requisition.height;
		else
			menu_y += widget->allocation.height;
	}
	else
	{
		if (pointer_y < widget->allocation.height &&
		    requisition.height < pointer_y)
			menu_y +=
			MIN (pointer_y,
			     widget->allocation.height -
			     requisition.height);
		menu_y = MIN (menu_y,
		              gdk_screen_get_height (screen) -
		              requisition.height);

		if (menu_x > gdk_screen_get_width (screen) / 2)
			menu_x -= requisition.width;
		else
			menu_x += widget->allocation.width;

	}

	*x = menu_x;
	*y = menu_y;
	*push_in = TRUE;
}

static void
talika_task_close_all (GtkMenuItem * menu_item, gpointer data)
{
	TalikaTask *task = data;
	GList *l;

	l = task->windows;
	while (l)
	{
		TalikaTask *child = l->data;
		wnck_window_close (child->window,
		                   gtk_get_current_event_time ());
		l = l->next;
	}
}

static void
talika_task_unminimize_all (GtkMenuItem * menu_item, gpointer data)
{
	TalikaTask *task = data;
	GList *l;

	l = task->windows;
	while (l)
	{
		TalikaTask *child = l->data;

		//update cache first
		talika_preview_image_cache_update_window (child->tasklist->preview_cache,
		                                          child->window,
		                                          child->tasklist->preview_size,
		                                          child->tasklist->box);
		
		/* This is inside an activate callback, so gtk_get_current_event_time()
		* will work.
		 */
		wnck_window_unminimize (child->window,
		                        gtk_get_current_event_time ());
		l = l->next;
	}
}

static void
talika_task_minimize_all (GtkMenuItem * menu_item, gpointer data)
{
	TalikaTask *task = data;
	GList *l;

	l = task->windows;
	while (l)
	{
		TalikaTask *child = l->data;

		//update cache first
		talika_preview_image_cache_update_window (child->tasklist->preview_cache,
		                                          child->window,
		                                          child->tasklist->preview_size,
		                                          child->tasklist->box);
		
		wnck_window_minimize (child->window);
		l = l->next;
	}
}

static void
talika_task_unmaximize_all (GtkMenuItem * menu_item, gpointer data)
{
	TalikaTask *task = data;
	GList *l;

	l = task->windows;
	while (l)
	{
		TalikaTask *child = l->data;
		wnck_window_unmaximize (child->window);
		l = l->next;
	}
}

static void
talika_task_maximize_all (GtkMenuItem * menu_item, gpointer data)
{
	TalikaTask *task = data;
	GList *l;

	l = task->windows;
	while (l)
	{
		TalikaTask *child = l->data;
		wnck_window_maximize (child->window);
		l = l->next;
	}
}

static void
talika_task_menu_activated (GtkMenuItem * menu_item, gpointer data)
{
	TalikaTask *task = data;

	/* This is an "activate" callback function so gtk_get_current_event_time()
	* will suffice.
	 */
	talika_tasklist_activate_task_window (task,
	                                      gtk_get_current_event_time (),FALSE);
}

static void 
talika_task_pin (GtkMenuItem *menu_item, TalikaTask *task)
{
	talika_tasklist_add_launcher (task);
}

static void
talika_task_add_launcher_menuitems (GtkWidget *menu, TalikaTask *task)
{
	GtkWidget *menu_item;
	GtkWidget *image;
	
	if(task->launcher)
	{
		menu_item = gtk_separator_menu_item_new();
		gtk_menu_shell_prepend (GTK_MENU_SHELL(menu), menu_item);

		menu_item = gtk_image_menu_item_new_with_mnemonic (_("Change icon"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_CHANGE_ICON,
		                                  GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);

		gtk_menu_shell_prepend (GTK_MENU_SHELL(menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item),
		                  "activate",
		                  G_CALLBACK (talika_launcher_change_icon), task->launcher);
		
		menu_item = gtk_image_menu_item_new_with_mnemonic (_("Launch"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_LAUNCH,
		                                  GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);
		
		gtk_menu_shell_prepend (GTK_MENU_SHELL(menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item),
		                  "activate",
		                  G_CALLBACK (talika_launcher_launch), task->launcher);

		
		menu_item = gtk_image_menu_item_new_with_mnemonic (_("Unpin"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_UNPIN,
		                                  GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);
		
		gtk_menu_shell_prepend (GTK_MENU_SHELL(menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item),
		                  "activate",
		                  G_CALLBACK (talika_launcher_unpin), task->launcher);
	}
	else 
	{
		WnckWindow *window=NULL;
		switch (task->type)
		{
			case TALIKA_TASK_WINDOW:
				window = task->window;
				break;

			case TALIKA_TASK_CLASS_GROUP:
				window = ((TalikaTask *)(task->windows->data))->window;
				break;

			default:
				break;
		}
		
		TalikaClass *class_obj = 
			talika_class_cache_get_class (task->tasklist->class_cache, window);

		if(class_obj->desktop_id)
		{
			menu_item = gtk_separator_menu_item_new();
			gtk_menu_shell_prepend (GTK_MENU_SHELL(menu), menu_item);
			
			menu_item = gtk_image_menu_item_new_with_mnemonic (_("Pin"));
			image = gtk_image_new_from_stock (TALIKA_STOCK_PIN,
			                                  GTK_ICON_SIZE_MENU);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
			                               (menu_item), image);
			
			gtk_menu_shell_prepend (GTK_MENU_SHELL(menu), menu_item);
			g_signal_connect (G_OBJECT (menu_item),
			                  "activate",
			                  G_CALLBACK (talika_task_pin), task);
		}
	}
}

static void
talika_task_popup_menu (TalikaTask * task, gboolean action_submenu)
{
	GtkWidget *menu;
	TalikaTask *win_task;
	char *text;
	GdkPixbuf *pixbuf;
	GtkWidget *menu_item;
	GtkWidget *image;
	GList *l, *list;

	g_return_if_fail (task->type == TALIKA_TASK_CLASS_GROUP);

	if (task->class_group == NULL)
		return;

	if (task->menu == NULL)
	{
		task->menu = gtk_menu_new ();
		g_object_ref_sink (task->menu);
	}

	menu = task->menu;

	/* Remove old menu content */
	list = gtk_container_get_children (GTK_CONTAINER (menu));
	l = list;
	while (l)
	{
		GtkWidget *child = GTK_WIDGET (l->data);
		gtk_container_remove (GTK_CONTAINER (menu), child);
		l = l->next;
	}
	g_list_free (list);

	l = task->windows;
	while (l)
	{
		win_task = l->data;

		text = talika_task_get_text (win_task);
		menu_item = gtk_image_menu_item_new_with_label (text);
		g_free (text);

		//gtk_image_menu_item_set_always_show_image (GTK_IMAGE_MENU_ITEM
		//                                         (menu_item), TRUE);

		if (talika_task_get_needs_attention (win_task))
			_make_gtk_label_bold (GTK_LABEL
			                      (GTK_BIN (menu_item)->child));

		text = talika_task_get_text (win_task);
		gtk_widget_set_tooltip_text (menu_item, text);
		g_free (text);

		pixbuf = gdk_pixbuf_copy (wnck_window_get_mini_icon
		                          (win_task->window));
		if (pixbuf)
		{
			image = gtk_image_new_from_pixbuf (pixbuf);
			gtk_widget_show (image);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
			                               (menu_item), image);
			g_object_unref (pixbuf);
		}

		gtk_widget_show (menu_item);

		if (action_submenu)
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item),
			                           wnck_action_menu_new
			                           (win_task->window));
		else
		{
			g_signal_connect (G_OBJECT (menu_item), "activate",
			                  G_CALLBACK
			                  (talika_task_menu_activated),
			                  win_task);
		}

		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);

		l = l->next;
	}

	/* In case of Right click, show Minimize All, Unminimize All, Close All */
	if (action_submenu)
	{
		GtkWidget *separator;
		GtkWidget *image;

		separator = gtk_separator_menu_item_new ();
		gtk_widget_show (separator);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), separator);

		menu_item =
			gtk_image_menu_item_new_with_mnemonic (_
			                                       ("Mi_nimize All"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_MINIMIZE,
		                                  GTK_ICON_SIZE_MENU);
		gtk_widget_show (image);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);
		gtk_widget_show (menu_item);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item), "activate",
		                  G_CALLBACK (talika_task_minimize_all),
		                  task);

		menu_item =
			gtk_image_menu_item_new_with_mnemonic (_
			                                       ("Un_minimize All"));
		gtk_widget_show (menu_item);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item), "activate",
		                  G_CALLBACK (talika_task_unminimize_all),
		                  task);

		menu_item =
			gtk_image_menu_item_new_with_mnemonic (_
			                                       ("Ma_ximize All"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_MAXIMIZE,
		                                  GTK_ICON_SIZE_MENU);
		gtk_widget_show (image);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);
		gtk_widget_show (menu_item);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item), "activate",
		                  G_CALLBACK (talika_task_maximize_all),
		                  task);

		menu_item =
			gtk_image_menu_item_new_with_mnemonic (_
			                                       ("_Unmaximize All"));
		gtk_widget_show (menu_item);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item), "activate",
		                  G_CALLBACK (talika_task_unmaximize_all),
		                  task);

		separator = gtk_separator_menu_item_new ();
		gtk_widget_show (separator);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), separator);

		menu_item =
			gtk_image_menu_item_new_with_mnemonic (_
			                                       ("_Close All"));
		image = gtk_image_new_from_stock (TALIKA_STOCK_CLOSE,
		                                  GTK_ICON_SIZE_MENU);
		gtk_widget_show (image);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM
		                               (menu_item), image);
		gtk_widget_show (menu_item);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item), "activate",
		                  G_CALLBACK (talika_task_close_all), task);

		talika_task_add_launcher_menuitems (menu, task);
	}

	gtk_menu_set_screen (GTK_MENU (menu),
	                     talika_screen_get_gdk_screen (task->
	                                                   tasklist->screen));

	gtk_widget_show_all (menu);
	gtk_menu_popup (GTK_MENU (menu),
	                NULL, NULL,
	                talika_task_position_menu, task,
	                1, gtk_get_current_event_time ());
}

static gboolean
talika_task_button_press_event (GtkWidget * widget,
                                GdkEventButton * event, gpointer data)
{
	TalikaTask *task = data;

	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
		{
			if (event->button == 1)
			{
				talika_tooltips_hide_popup_immediatly (task);
				return FALSE;
			}
			else if (event->button == 2)
			{
				if(task->tasklist->group_middle_close)
				{
					talika_task_close_all (NULL,task);
				}
			}
			else if (event->button == 3)
			{
				talika_tooltips_hide_popup_immediatly (task);
				talika_task_popup_menu (task, TRUE);
			}
			return TRUE;
		}

		case TALIKA_TASK_WINDOW:
		{
			if (event->button == 1)
			{

				/* is_most_recently_activated == is_active for click &
				 * sloppy focus methods.  We use the former here because
				 * 'mouse' focus provides a special case.  In that case, no
				 * window will be active, but if a window was the most
				 * recently active one (i.e. user moves mouse straight from
				                        * window to tasklist), then we should still minimize it.
											*/
				if (wnck_window_is_most_recently_activated
				    (task->window))
					task->was_active = TRUE;
				else
					task->was_active = FALSE;

				return FALSE;
			}
			else if (event->button == 2)
			{
				talika_tooltips_hide_popup_immediatly (task);
				wnck_window_close (task->window,
				                   gtk_get_current_event_time ());
			}
			else if (event->button == 3)
			{
				talika_tooltips_hide_popup_immediatly (task);

				if (task->action_menu)
					gtk_widget_destroy (task->action_menu);

				g_assert (task->action_menu == NULL);

				task->action_menu =
					wnck_action_menu_new (task->window);

				talika_task_add_launcher_menuitems (task->action_menu, task);

				g_object_add_weak_pointer (G_OBJECT
				                           (task->action_menu),
				                           (void **)
				                           &task->action_menu);

				gtk_menu_set_screen (GTK_MENU (task->action_menu),
				                     talika_screen_get_gdk_screen
				                     (task->tasklist->screen));

				gtk_widget_show_all (task->action_menu);
				gtk_menu_popup (GTK_MENU (task->action_menu),
				                NULL, NULL,
				                talika_task_position_menu,
				                task, event->button,
				                gtk_get_current_event_time ());

				g_signal_connect (task->action_menu, "selection-done",
				                  G_CALLBACK (gtk_widget_destroy),
				                  NULL);

				return TRUE;
			}
		}
	}
	return FALSE;
}

static void
talika_task_button_toggled (GtkButton * button, TalikaTask * task)
{
	/* Did we really want to change the state of the togglebutton? */
	if (task->really_toggling)
		return;

	/* Undo the toggle */
	task->really_toggling = TRUE;
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
	                              !gtk_toggle_button_get_active
	                              (GTK_TOGGLE_BUTTON (button)));
	task->really_toggling = FALSE;

	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
		{
			gboolean preview_enabled = task->tasklist->enable_previews &&
				gdk_screen_is_composited (gtk_widget_get_screen(task->button));
			
			if(preview_enabled && task->tasklist->previews_group_click)
			{
				talika_tooltips_show_popup_immediatly(task);
			}
			else
				talika_task_popup_menu (task, FALSE);
		}
		break;

		case TALIKA_TASK_WINDOW:
			if (task->window == NULL)
				return;

			/* This should only be called by clicking on the task button, so
			 * gtk_get_current_event_time() should be fine here...
				 */
			talika_tasklist_activate_task_window (task,
			                                      gtk_get_current_event_time(),FALSE);
			break;
	}
}


/*########################## Manage Button Icons #############################*/
GdkPixbuf*
talika_desaturate_icon (GdkPixbuf *pixbuf)
{
	GdkPixbuf *desat = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
                            TRUE,
                            gdk_pixbuf_get_bits_per_sample (pixbuf),
                            gdk_pixbuf_get_width (pixbuf),
                            gdk_pixbuf_get_height (pixbuf));

	gdk_pixbuf_saturate_and_pixelate (pixbuf,
                                        desat,
                                        0,
                                       FALSE);

	return desat;
}

static GdkPixbuf*
talika_dimm_icon (GdkPixbuf *pixbuf)
{
	int x, y, pixel_stride, row_stride;
	guchar *row, *pixels;
	int w, h;

	//talika exits on startup sometime because icon is null
	//checking here will handle all cases
	if(pixbuf == NULL)
		return NULL;

	w = gdk_pixbuf_get_width (pixbuf);
	h = gdk_pixbuf_get_height (pixbuf);

	if (!gdk_pixbuf_get_has_alpha (pixbuf))
	{
		GdkPixbuf *temp =
			gdk_pixbuf_add_alpha (pixbuf, FALSE, 0, 0, 0);
		g_object_unref (pixbuf);
		pixbuf = temp;
	}
	else
	{
		GdkPixbuf *temp =
			gdk_pixbuf_copy (pixbuf);
		g_object_unref (pixbuf);
		pixbuf = temp;
	}

	pixel_stride = 4;

	row = gdk_pixbuf_get_pixels (pixbuf);
	row_stride = gdk_pixbuf_get_rowstride (pixbuf);

	for (y = 0; y < h; y++)
	{
		pixels = row;

		for (x = 0; x < w; x++)
		{
			pixels[3] /= 2;

			pixels += pixel_stride;
		}

		row += row_stride;
	}

	return pixbuf;
}

/*static GdkPixbuf *
talika_get_installed_app_icon (TalikaTask * task, guint size)
{
	GError *err = NULL;
	GAppInfo *app = g_app_info_create_from_commandline ("gimp",NULL,
	                                                    G_APP_INFO_CREATE_NONE,
	                                                    &err);

	if(err)
	{
		g_print("error\n");
	}

	g_print("%s",g_app_info_get_description(app));
	g_print("\n");
	GIcon *icon = g_app_info_get_icon (app);
	if(!icon)
	{
		g_print("no icon\n");
	}
	//g_print( g_icon_to_string (icon));
	
	return NULL;
}
*/

static GdkPixbuf *
talika_get_hq_icon (TalikaTask * task, guint size)
{
	if (task->type == TALIKA_TASK_WINDOW)
		return talika_get_icon_from_xserver (task->window, size);

	return NULL;
}

static GdkPixbuf *
talika_sized_pixbuf_for_task (TalikaTask * task, gint size)
{
	GdkPixbuf *pbuf = NULL;

	if (task->type == TALIKA_TASK_WINDOW)
		pbuf = gdk_pixbuf_copy (wnck_window_get_icon (task->window));
	else
		pbuf = gdk_pixbuf_copy (wnck_class_group_get_icon
		                        (task->class_group));

	gint width = gdk_pixbuf_get_width (pbuf);
	gint height = gdk_pixbuf_get_height (pbuf);

	if (MAX (width, height) != size)
	{
		gdouble scale =
			(gdouble) size / (gdouble) MAX (width, height);

		GdkPixbuf *tmp = pbuf;
		pbuf = gdk_pixbuf_scale_simple (tmp, (gint) (width * scale),
		                                (gint) (height * scale),
		                                GDK_INTERP_HYPER);

		g_object_unref (tmp);
	}

	return pbuf;
}

static GdkPixbuf *
talika_task_get_window_icon (TalikaTask * task)
{
	GdkPixbuf *pixbuf = NULL;

	guint size = task->tasklist->panel_size;
	size -= task->tasklist->icon_size_offset;		//subtract box and button paddings

	//show launcher icon 
	if (task->launcher && task->tasklist->show_theme_icons)
	{
		pixbuf = talika_launcher_get_icon (task->launcher);
	}
	
	//first look if theme has icon, dimm icon may fail 
	if (!pixbuf && task->tasklist->show_theme_icons)
	{
		TalikaClass *class_obj = 
			talika_class_cache_get_class (task->tasklist->class_cache,
			                              task->window);
		
		if(class_obj)
			pixbuf = 
			talika_get_icon_from_desktop_file (task->tasklist,class_obj->desktop_id,size);
	}

	if (!pixbuf)
	{
		//second ask x11 for icon
		pixbuf = talika_get_hq_icon (task, size);

		//finaly do the trick
		if (!pixbuf)
		{
			//g_print ("from wnck : %s\n",
			  //       wnck_window_get_name (task->window));

			if (size >= 16 && size < 19)
			{
				pixbuf = gdk_pixbuf_copy
					(wnck_window_get_mini_icon
					 (task->window));
			}
			else if (size > 30 && size < 36)
			{
				pixbuf = gdk_pixbuf_copy (wnck_window_get_icon
				                          (task->window));
			}
			else
				pixbuf = talika_sized_pixbuf_for_task
				(task, size);
		}
		else
		{
			//g_print ("from xwin : %s\n",
			  //       wnck_window_get_name (task->window));
		}
	}
	else
	{
		//g_print ("from theme : %s\n",
		 //        wnck_window_get_name (task->window));
	}

	if (task->tasklist->dimm_minimised &&
	    wnck_window_is_minimized (task->window))
	{
		pixbuf = talika_dimm_icon (pixbuf);
	}
	
	return pixbuf;
}

static GdkPixbuf *
talika_task_get_class_group_icon (TalikaTask * task)
{
	GdkPixbuf *pixbuf = NULL;

	guint size = task->tasklist->panel_size;
	size -= task->tasklist->icon_size_offset;		//subtract box and button paddings

	//show launcher icon 
	if (task->launcher)
	{
		pixbuf = talika_launcher_get_icon (task->launcher);
	}
	
	//first look if a desktop file exists
	if (!pixbuf && task->windows)
	{
		WnckWindow *window = ((TalikaTask *)(task->windows->data))->window;
		
		TalikaClass *class_obj = 
			talika_class_cache_get_class (task->tasklist->class_cache, window);
		
		if(class_obj)
			pixbuf = 
			talika_get_icon_from_desktop_file (task->tasklist,class_obj->desktop_id,size);
	}
	
	//finaly do the trick
	if (!pixbuf)
	{
		//g_print ("from wnck : %s\n",
		//         wnck_class_group_get_name (task->class_group));

		if (size >= 16 && size < 19)
		{
			pixbuf = gdk_pixbuf_copy
				(wnck_class_group_get_mini_icon
				 (task->class_group));
		}
		else if (size > 30 && size < 36)
		{
			pixbuf = gdk_pixbuf_copy (wnck_class_group_get_icon
			                          (task->class_group));
		}
		else
			pixbuf = talika_sized_pixbuf_for_task (task, size);
	}
	else
	{
		//g_print ("from theme : %s\n",
		//         wnck_class_group_get_name (task->class_group));
	}

	GList *l;
	gboolean all_minimised = TRUE;
	l = task->windows;
	while (l != NULL)
	{
		TalikaTask *task1 = (TalikaTask *) l->data;
		l = l->next;

		if (!wnck_window_is_minimized (task1->window))
		{
			all_minimised = FALSE;
			break;
		}
	}

	if (task->tasklist->dimm_minimised && all_minimised)
	{
		pixbuf = talika_dimm_icon (pixbuf);
	}


	return pixbuf;
}

GdkPixbuf*
talika_task_get_icon (TalikaTask * task)
{
	GdkPixbuf *pixbuf = NULL;

	/*guint size = task->tasklist->panel_size - task->tasklist->icon_size_offset;	
		pixbuf = gdk_pixbuf_new_from_file_at_size ("/usr/share/pixmaps/apple-red.png",
		                                           size,
		                                           size,
		                                           NULL);
	return pixbuf;*/
	
	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
			pixbuf = talika_task_get_class_group_icon (task);
			break;

		case TALIKA_TASK_WINDOW:
			pixbuf = talika_task_get_window_icon (task);
			break;

		case TALIKA_TASK_LAUNCHER:
			pixbuf = talika_launcher_get_icon (task);
			break;
	}

	if(task->tasklist->mono_icons && pixbuf)
	{
		GdkPixbuf *temp = pixbuf;
		pixbuf = talika_desaturate_icon (temp);
		g_object_unref (temp);
	}

	if(!pixbuf)
	{
		pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB,TRUE,8,
		                         task->tasklist->panel_size -task->tasklist->icon_size_offset,
		                         task->tasklist->panel_size -task->tasklist->icon_size_offset);
	}
	
	return pixbuf;
}

gchar*
talika_task_get_text (TalikaTask * task)
{
	const char *name;

	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
			if (task->windows)
			{
				WnckWindow *window = ((TalikaTask *)(task->windows->data))->window;

				TalikaClass *class_obj = 
					talika_class_cache_get_class (task->tasklist->class_cache, window);

				name = class_obj->class_name;
			}
			else
				name = wnck_class_group_get_name (task->class_group);
			
			if (name[0] != 0)
				return g_strdup_printf ("%s (%d)",
				                        name,
				                        g_list_length
				                        (task->windows));
			else
				return g_strdup_printf ("(%d)",
				                        g_list_length
				                        (task->windows));

		case TALIKA_TASK_WINDOW:
			name = wnck_window_get_name (task->window);

			if (name)
				return g_strdup (name);

			break;
			
		case TALIKA_TASK_LAUNCHER:
		{
			gchar *name2 = g_strdup_printf(_("Launch %s"),
	                              g_app_info_get_name 
	                              (G_APP_INFO(task->app_info)));

			if (name2)
				return name2;
		}
		break;
	}

	return NULL;
}

/*###################### Task proprty change Callbacks #######################*/

void 
talika_task_update_visible_state (TalikaTask * task)
{
	GdkPixbuf *pixbuf;

	gtk_button_set_relief (GTK_BUTTON (task->button),
	                       task->
	                       tasklist->button_relief ? GTK_RELIEF_NORMAL :
		                       GTK_RELIEF_NONE);

	pixbuf = talika_task_get_icon (task);
	button_widget_set_pixbuf (BUTTON_WIDGET(task->button),pixbuf);
	g_object_unref (pixbuf);

	if (talika_task_get_needs_attention (task))
	{
		talika_task_queue_glow (task);
	}
	else
	{
		talika_task_stop_glow (task);
	}

	//gtk_widget_queue_resize (GTK_WIDGET (task->tasklist->box));
}

void 
do_talika_task_update_visible_state (gpointer data, 
                                     gpointer user_data)
{
	TalikaTask *task = (TalikaTask *) data;
	talika_task_update_visible_state (task);
}

static gboolean
talika_task_update_preview_cache_timeout (gpointer data)
{
	TalikaTask *task = (TalikaTask *) data;

	task->preview_update_timeout_id = 0;
	
	//update cache
	talika_preview_image_cache_update_window (task->tasklist->preview_cache,
	                                          task->window,
	                                          task->tasklist->preview_size,
	                                          task->tasklist->box);

	return FALSE;
}

void 
talika_task_state_changed (WnckWindow * window,
                           WnckWindowState changed_mask,
                           WnckWindowState new_state, gpointer data)
{
	TalikaTaskList *tasklist = data;

	//state changed, update cache
	if (!(new_state & WNCK_WINDOW_STATE_MINIMIZED) &&
	    !(new_state & WNCK_WINDOW_STATE_SKIP_TASKLIST))
	{
		TalikaTask *task = NULL;
		WnckWindow *win_temp = window;
		// FIXME: Handle group modal dialogs 
		for (; win_temp && !task;
		     win_temp = wnck_window_get_transient (win_temp))
			task =
			g_hash_table_lookup (tasklist->win_hash,
			                     win_temp);

		if(task)
		{
			if (task->preview_update_timeout_id)
				g_source_remove (task->preview_update_timeout_id);

			task->preview_update_timeout_id =
				g_timeout_add (500, &talika_task_update_preview_cache_timeout,
				               task);
		}
	}
	///////////////////////////////

	if (changed_mask & WNCK_WINDOW_STATE_SKIP_TASKLIST)
	{
		talika_tasklist_update_lists (tasklist);
		//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
		return;
	}

	if ((changed_mask & WNCK_WINDOW_STATE_DEMANDS_ATTENTION) ||
	    (changed_mask & WNCK_WINDOW_STATE_URGENT))
	{
		WnckWorkspace *active_workspace =
			wnck_screen_get_active_workspace (tasklist->screen);

		if (active_workspace &&
		    (active_workspace !=
		     wnck_window_get_workspace (window)
		     || (wnck_workspace_is_virtual (active_workspace)
		         && !wnck_window_is_in_viewport (window,
		                                         active_workspace))))
		{
			talika_tasklist_update_lists (tasklist);
			//gtk_widget_queue_resize (GTK_WIDGET (tasklist->box));
		}
	}

	if ((changed_mask & WNCK_WINDOW_STATE_MINIMIZED) ||
	    (changed_mask & WNCK_WINDOW_STATE_DEMANDS_ATTENTION) ||
	    (changed_mask & WNCK_WINDOW_STATE_URGENT))
	{
		TalikaTask *win_task = NULL;

		// FIXME: Handle group modal dialogs 
		for (; window && !win_task;
		     window = wnck_window_get_transient (window))
			win_task =
			g_hash_table_lookup (tasklist->win_hash,
			                     window);

		if (win_task)
		{
			talika_task_update_visible_state (win_task);

			TalikaTask *class_group_task;

			class_group_task =
				g_hash_table_lookup
				(tasklist->class_group_hash,
				 win_task->class_group_name);

			if (class_group_task)
				talika_task_update_visible_state
				(class_group_task);
		}
	}
}

static void
talika_task_icon_changed (WnckWindow * window, gpointer data)
{
	TalikaTask *task = data;

	if (task)
		talika_task_update_visible_state (task);
}

static void
talika_task_name_changed (WnckWindow * window, gpointer data)
{
	TalikaTask *task = data;

	if (task)
		talika_task_update_visible_state (task);

	//fix openoffice desktop id reconigtion bug
	TalikaClass *class_obj = 
		talika_class_cache_get_class (task->tasklist->class_cache,
		                              window);

	if(class_obj)
	{
		if(g_ascii_strncasecmp (class_obj->class_name,"OpenOffice.org",14)==0
		   && class_obj->desktop_id == NULL)
		{
			talika_class_cache_update_window (task->tasklist->class_cache,
			                                  window);
			talika_tasklist_update_lists (task->tasklist);
		}
	}
}

static void
talika_task_class_name_changed (WnckClassGroup * class_group, gpointer data)
{
	TalikaTask *task = data;

	if (task)
		talika_task_update_visible_state (task);
}

static void
talika_task_class_icon_changed (WnckClassGroup * class_group, gpointer data)
{
	TalikaTask *task = data;

	if (task)
		talika_task_update_visible_state (task);
}

/*################################ Task ######################################*/

static void
talika_task_create_widgets (TalikaTask * task)
{
	//GdkPixbuf *pixbuf;

	static const GtkTargetEntry targets[] = {
		{"application/x-wnck-window-id", 0, 0},
		{"text/uri-list", 0, 1}
	};

	task->button = button_widget_new (NULL,task);

	g_object_add_weak_pointer (G_OBJECT (task->button),
	                           (void **) &task->button);

	gtk_widget_set_name (task->button, "tasklist-button");


	gtk_drag_source_set (GTK_WIDGET (task->button),
	                     GDK_BUTTON1_MASK, targets, 1, GDK_ACTION_MOVE);
	gtk_drag_dest_set (GTK_WIDGET (task->button),
	                   GTK_DEST_DEFAULT_DROP, targets, 2,
	                   GDK_ACTION_MOVE);

	gtk_button_set_relief (GTK_BUTTON (task->button),
	                       task->tasklist->button_relief ?
	                       GTK_RELIEF_NORMAL : GTK_RELIEF_NONE);

	/*//some gtk styles have big button thickness, it tends to
	//distort either button or image
	//change button style here
	{
		GtkRcStyle *mod;

		mod = gtk_widget_get_modifier_style (task->button);
		mod->xthickness = 1;
		mod->ythickness = 1;

		gtk_widget_modify_style (task->button, mod);
	}*/

	//pixbuf = talika_task_get_icon (task);
	//button_widget_set_pixbuf (BUTTON_WIDGET(task->button),pixbuf);
	//g_object_unref (pixbuf);

	gtk_widget_show_all (task->button);


	if (talika_task_get_needs_attention (task))
	{
		talika_task_queue_glow (task);
	}

	g_signal_connect (G_OBJECT (task->button),
	                  "button_press_event",
	                  G_CALLBACK (talika_task_button_press_event), task);

	g_signal_connect (G_OBJECT (task->button), "toggled",
	                  G_CALLBACK (talika_task_button_toggled), task);

	int i = 0;
	switch (task->type)
	{
		case TALIKA_TASK_CLASS_GROUP:
			task->class_connections[i++] =
				g_signal_connect (G_OBJECT (task->class_group),
				                  "name_changed",
				                  G_CALLBACK
				                  (talika_task_class_name_changed),
				                  task);
			task->class_connections[i++] =
				g_signal_connect (G_OBJECT (task->class_group),
				                  "icon_changed",
				                  G_CALLBACK
				                  (talika_task_class_icon_changed),
				                  task);
			break;

		case TALIKA_TASK_WINDOW:
			i = 0;
			task->window_connections[i++] =
				g_signal_connect (G_OBJECT (task->window),
				                  "state_changed",
				                  G_CALLBACK
				                  (talika_task_state_changed),
				                  task->tasklist);
			task->window_connections[i++] =
				g_signal_connect (G_OBJECT (task->window),
				                  "icon_changed",
				                  G_CALLBACK
				                  (talika_task_icon_changed), task);
			task->window_connections[i++] =
				g_signal_connect (G_OBJECT (task->window),
				                  "name_changed",
				                  G_CALLBACK
				                  (talika_task_name_changed), task);
			break;
	}

	g_signal_connect (G_OBJECT (task->button), "drag_motion",
	                  G_CALLBACK (talika_task_drag_motion), task);


	g_signal_connect (G_OBJECT (task->button),
	                  "drag_data_get",
	                  G_CALLBACK (talika_task_drag_data_get), task);

	g_signal_connect (G_OBJECT (task->button),
	                  "drag_data_received",
	                  G_CALLBACK (talika_task_drag_data_received), task);

	g_signal_connect (G_OBJECT (task->button),
	                  "drag_begin",
	                  G_CALLBACK (talika_task_drag_begin), task);

	g_signal_connect (G_OBJECT (task->button), "drag_end",
	                  G_CALLBACK (talika_task_drag_end), task);


	g_signal_connect (G_OBJECT (task->button), "drag_leave",
	                  G_CALLBACK (talika_task_drag_leave), task);


	g_signal_connect_after (task->button, "expose_event",
	                        G_CALLBACK (talika_task_expose), task);

	g_signal_connect_after (task->button, "enter-notify-event",
	                        G_CALLBACK (talika_tooltips_enter_notify), task);

	g_signal_connect_after (task->button, "leave-notify-event",
	                        G_CALLBACK (talika_tooltips_leave_notify),
	                        task->tasklist);

}

TalikaTask* 
talika_task_new_from_window (TalikaTaskList * tasklist,
                             WnckWindow * window,
                             const gchar *class_group_name)
{
	TalikaTask *task;

	task = g_new0 (TalikaTask, 1);

	task->type = TALIKA_TASK_WINDOW;
	task->window = window;
	task->class_group = wnck_window_get_class_group (window);
	task->class_group_name = g_strdup (class_group_name);
	task->tasklist = tasklist;
	task->launcher = NULL;

	talika_task_create_widgets (task);
	return task;
}

TalikaTask* 
talika_task_new_from_class_group (TalikaTaskList * tasklist,
                                  WnckClassGroup * class_group,
                                  const gchar *class_group_name)
{
	TalikaTask *task;

	task = g_new0 (TalikaTask, 1);

	task->type = TALIKA_TASK_CLASS_GROUP;
	task->window = NULL;
	task->class_group = class_group;
	task->class_group_name = g_strdup (class_group_name);
	task->tasklist = tasklist;
	task->launcher = NULL;

	talika_task_create_widgets (task);
	return task;
}

void 
talika_task_free (TalikaTask * task)
{
	talika_tooltips_hide_popup_immediatly (task);
	
	if (task->tasklist->active_task == task)
		talika_tasklist_change_active_task (task->tasklist, NULL);

	if (task->button)
	{
		g_object_remove_weak_pointer (G_OBJECT (task->button),
		                              (void **) &task->button);
		gtk_widget_destroy (task->button);
		task->button = NULL;
	}

	//disconnect signals
	int i = 0;
	while (i < N_WINDOW_CONNECTIONS)
	{
		if (task->window_connections[i] != 0)
			g_signal_handler_disconnect (G_OBJECT
			                             (task->window),
			                             task->window_connections
			                             [i]);
		task->window_connections[i] = 0;
		++i;
	}

	if (task->menu)
	{
		gtk_widget_destroy (task->menu);
		task->menu = NULL;
	}

	if (task->action_menu)
	{
		g_object_remove_weak_pointer (G_OBJECT
		                              (task->action_menu),
		                              (void **) &task->action_menu);
		gtk_widget_destroy (task->action_menu);
		task->action_menu = NULL;
	}

	if (task->drop_popup)
	{
		gtk_widget_destroy (task->drop_popup);
		task->drop_popup = NULL;
	}

	task->window = NULL;

	g_list_free (task->windows);
	task->windows = NULL;

	//disconnect class connections
	i = 0;
	while (i < N_CLASS_CONNECTIONS)
	{
		if (task->class_connections[i] != 0)
			g_signal_handler_disconnect (G_OBJECT
			                             (task->class_group),
			                             task->class_connections
			                             [i]);
		task->class_connections[i] = 0;
		++i;
	}

	talika_task_stop_glow (task);

	cleanup_screenshots (task);

	if (task->preview_update_timeout_id != 0)
	{
		g_source_remove (task->preview_update_timeout_id);
		task->preview_update_timeout_id = 0;
	}

	g_free (task->class_group_name);

	g_free (task);
	task=NULL;
}
